@@ -4,7 +4,10 @@
  * Concept due to Harry Hochheiser.  Implementation by ESR.  Cleanup and
  * strict RFC821 compliance by Cameron MacPherson.
  *
- * Copyright 1997 Eric S. Raymond, 2009 - 2025 Matthias Andree
+ * Relevant standards: RFC-5321, for SMTP AUTH: RFC-4954, RFC-4422.
+ *
+ * Copyright 1997 Eric S. Raymond
+ * Copyright 2009 - 2025 Matthias Andree
  * Contribution 2004 by Phil Endecott (by way of Rob Funk)
  * Contributions 2005, 2011 by Sunil Shetye
  * Contributions 2012, 2021 by Earl Chew
@@ -15,6 +18,7 @@
 #include "fetchmail.h"
 
 #include <ctype.h>
+#include <errno.h>
 #include <limits.h>
 #include <stdio.h>
 #include <unistd.h>
@@ -87,9 +91,31 @@ static void SMTP_auth_error(int sock, const char *msg /** Message to print/log i
     if (outlevel >= O_MONITOR && msg && *msg) report(stdout, "%s", msg);
 }
 
+/** This function checks if the \a smtp_reply starts with "334 " and reports either "Server rejected the AUTH command" (if the first part isn't 334)
+or "Malformed server reply" if the blank is missing.
+\returns
+  - true if the reply starts with "334 "
+  - false if the reply does not start with "334 ". */
+static bool SMTP_check_AUTH_reply(const char *smtp_reply /** input: full reply line received from SMTP server */)
+{
+  if (strncmp(smtp_reply, "334", 3)) { /* Server rejects AUTH */
+    report(stderr, "\"%s\" <- %s", visbuf(smtp_reply),
+           GT_("Server rejected the AUTH command.\n"));
+    return false;
+  }
+  if (smtp_reply[3] != ' ') {
+    report(stderr, "\"%s\" <- %s",
+		    visbuf(smtp_reply),
+		    // GT_("Malformed server reply\n"),
+		    GT_("protocol error\n"));
+    return false;
+  }
+  return true;
+}
+
 /** Authenticate \a sock via ESMTP, currently supported mechanisms: CRAM-MD5, LOGIN, PLAIN.
- * Original ESMTP Authentication support for fetchmail by Wojciech Polak */
-static void SMTP_auth(int sock, char smtp_mode, const char *username, const char *password,
+ * Original ESMTP Authentication support for fetchmail by Wojciech Polak. */
+static int SMTP_auth(int sock, char smtp_mode, const char *username, const char *password,
         const char *ehlo_advertised_auth)
 {
 	int c;
@@ -97,17 +123,21 @@ static void SMTP_auth(int sock, char smtp_mode, const char *username, const char
 	char b64buf[512];
 	char tmp[512];
 
-	if (!username || !password) return;
+	if (!username || !password) {
+                report(stderr, "SMTP AUTH: %s\n", GT_("Invalid userid or passphrase"));
+                return SM_UNRECOVERABLE;
+        }
 
-        if (strlen(username) > INT_MAX) return;
+        if (strlen(username) > INT_MAX) return SM_UNRECOVERABLE;
         const int unlen = (int)strlen(username);
-        if (strlen(password) > INT_MAX) return;
+        if (strlen(password) > INT_MAX) return SM_UNRECOVERABLE;
         const int pwlen = (int)strlen(password);
 
 	memset(b64buf, 0, sizeof(b64buf));
 	memset(tmp, 0, sizeof(tmp));
 
 	if (strstr(ehlo_advertised_auth, "CRAM-MD5")) {
+                // IETF RFC 4952, 2195
 		unsigned char digest[16];
 		memset(digest, 0, sizeof(digest));
 
@@ -117,41 +147,52 @@ static void SMTP_auth(int sock, char smtp_mode, const char *username, const char
 		SockRead(sock, smtp_response, sizeof(smtp_response) - 1);
 		strlcpy(tmp, smtp_response, sizeof(tmp));
 
+                // could use SMTP_check_AUTH_reply() here instead, but we need the pointer "after the blank" also
+
 		if (strncmp(tmp, "334", 3)) { /* Server rejects AUTH */
 			report(stderr, "\"%s\" <- %s", visbuf(tmp), GT_("Server rejected the AUTH command.\n"));
-			SMTP_auth_error(sock, "");
-			return;
+                        // No need to cancel SMTP AUTH here.
+			return SM_ERROR;
 		}
 
 		p = strchr(tmp, ' ');
 		if (!p) {
 			report(stderr, "%s: \"%s\"\n", GT_("Malformed server reply"), visbuf(tmp));
 			SMTP_auth_error(sock, "");
-			return;
+			return SM_ERROR;
 		}
 		p++;
 		/* (hmh) from64tobits will not NULL-terminate strings! */
+                // The challenge cannot be empty, so we use "<= 0"
 		if (from64tobits(b64buf, p, sizeof(b64buf) - 1) <= 0) {
 			report(stderr, "\"%s\" <- %s", visbuf(tmp), GT_("Bad base64 reply from server.\n"));
 			SMTP_auth_error(sock, "");
-			return;
+			return SM_ERROR;
 		}
 		if (outlevel >= O_DEBUG)
 			report(stdout, GT_("Challenge decoded: %s\n"), b64buf);
 		hmac_md5((unsigned const char *)password, pwlen,
 			 (unsigned char *)b64buf, strlen(b64buf), digest, sizeof(digest));
-		snprintf(tmp, sizeof(tmp),
-		"%s %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
-		username,  digest[0], digest[1], digest[2], digest[3],
-		digest[4], digest[5], digest[6], digest[7], digest[8],
-		digest[9], digest[10], digest[11], digest[12], digest[13],
-		digest[14], digest[15]);
+		int plen = snprintf(tmp, sizeof(tmp),
+			"%s %02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x%02x",
+			username,  digest[0], digest[1], digest[2], digest[3],
+			digest[4], digest[5], digest[6], digest[7], digest[8],
+			digest[9], digest[10], digest[11], digest[12], digest[13],
+			digest[14], digest[15]);
+		if (plen < 0 || (size_t)plen >= sizeof(tmp))
+                {
+                        // username pretty long so we truncated the output - this is pointless to send, abort authentication instead
+                        report(stderr, GT_("Digest text buffer too small!\n")); // XXX FIXME: this isn't accurate but already translated.
+                        SMTP_auth_error(sock, "");
+                        return SM_ERROR;
+                }
 
 		to64frombits(b64buf, tmp, strlen(tmp), sizeof b64buf); // NOSONAR: tmp is our own buffer, so strlen() shan't exceed its size
 		SockPrintf(sock, "%s\r\n", b64buf);
-		SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
+		return SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
 	}
 	else if (strstr(ehlo_advertised_auth, "PLAIN")) {
+                // IETF RFC 4952, 4616.
 		int len;
 		if (outlevel >= O_MONITOR)
 			report(stdout, GT_("ESMTP PLAIN Authentication...\n"));
@@ -170,56 +211,62 @@ static void SMTP_auth(int sock, char smtp_mode, const char *username, const char
 		fm_safe_clearmem(tmp, sizeof(tmp));
 		SockPrintfClear(sock, "AUTH PLAIN %s\r\n", b64buf);
 		fm_safe_clearmem(b64buf, sizeof(b64buf));
-		SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
+		return SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
 	}
 	else if (strstr(ehlo_advertised_auth, "LOGIN")) {
+                // NONSTANDARD and also inefficient because it takes too many round trips
+                // compared to RFC-4616 PLAIN authentication, see above
+                // https://datatracker.ietf.org/doc/html/draft-murchison-sasl-login-00
+		// has expired and never became even a RFC, let alone a standard.
 		if (outlevel >= O_MONITOR)
 			report(stdout, GT_("ESMTP LOGIN Authentication...\n"));
+
 		SockPrintf(sock, "AUTH LOGIN\r\n");
 		SockRead(sock, smtp_response, sizeof(smtp_response) - 1);
 		strlcpy(tmp, smtp_response, sizeof(tmp));
 
-		if (strncmp(tmp, "334", 3)) { /* Server rejects AUTH */
-			report(stderr, "\"%s\" <- %s", visbuf(tmp), GT_("Server rejected the AUTH command.\n"));
-			SMTP_auth_error(sock, "");
-			return;
-		}
+                /* we used to check the base64-encoded username and password prompts.
+                   We don't do that any more, although RFC 4954 requires us to.
 
-		p = strchr(tmp, ' ');
-		if (!p) {
-			report(stderr, "%s: \"%s\"\n", GT_("Malformed server reply"), visbuf(tmp));
-			SMTP_auth_error(sock, "");
-			return;
-		}
-		p++;
-		if (from64tobits(b64buf, p, sizeof(b64buf) - 1) <= 0) {
-			report(stderr, "\"%s\" <- %s", visbuf(tmp), GT_("Bad base64 reply from server.\n"));
-			SMTP_auth_error(sock, "");
-			return;
-		}
-		to64frombits(b64buf, username, unlen, sizeof b64buf);
+                   1. If I understand recent bug fixes to mutt 2.2.15 properly, then
+                   some versions of Microsoft Exchange would send garbage here
+                   such as "GSSAPI supported" that isn't base64.
+                   <https://gitlab.com/muttmua/mutt/-/commit/b363b602e17906f63f3177f9899cd6c150e746ab>
+                   Quote "Another workaround for broken SMTP servers.  Instead of an empty challenge,
+                   some MS servers return a meaningless non-BASE64 encoded response in the initial reply,
+                   e.g. "334 GSSAPI supported".
+"
+                   2. Servers that comply with RFC-4952 must just send "334 \r\n" which our base64 decoder
+                   decodes to 0 bytes, and the former smtp.c client code required a non-zero length.
+                   If the server's reply starts with "334 ", we're good and continue.
+                 */
+                if (!SMTP_check_AUTH_reply(smtp_response)) {
+			// No need to cancel SMTP AUTH here.
+                        return SM_ERROR;
+                }
+
+                // send base64(username)
+                to64frombits(b64buf, username, unlen, sizeof b64buf);
 		SockPrintf(sock, "%s\r\n", b64buf);
 		SockRead(sock, smtp_response, sizeof(smtp_response) - 1);
-		strlcpy(tmp, smtp_response, sizeof(tmp));
-		p = strchr(tmp, ' ');
-		if (!p) {
-			report(stderr, "\"%s\" <- %s", visbuf(tmp), GT_("Bad base64 reply from server.\n"));
-			SMTP_auth_error(sock, "");
-			return;
-		}
-		p++;
-		memset(b64buf, 0, sizeof(b64buf));
-		if (from64tobits(b64buf, p, sizeof(b64buf) - 1) <= 0) {
-			report(stderr, "\"%s\" <- %s", visbuf(tmp), GT_("Bad base64 reply from server.\n"));
-			SMTP_auth_error(sock, "");
-			return;
-		}
+
+                if (!SMTP_check_AUTH_reply(smtp_response)) {
+                        return SM_ERROR;
+                }
+
+                /* again, RFC-4952 section 4 suggests we should decode the server
+                 * challenge and abort authentication if we can't decode it; again,
+                 * we just ignore anything after "334 ".
+                 */
+
+                // send base64(password)
 		to64frombits(b64buf, password, pwlen, sizeof b64buf);
 		SockPrintfClear(sock, "%s\r\n", b64buf);
 		fm_safe_clearmem(b64buf, sizeof(b64buf));
-		SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
+
+		return SMTP_ok(sock, smtp_mode, TIMEOUT_DEFAULT);
 	}
-	return;
+	return SM_UNRECOVERABLE;
 }
 
 int SMTP_ehlo(int sock, char smtp_mode /**< 'S' for ESMTP or 'L' for LMTP */,
@@ -261,8 +308,8 @@ int SMTP_ehlo(int sock, char smtp_mode /**< 'S' for ESMTP or 'L' for LMTP */,
 		strlcpy(auth_response, smtp_response, sizeof(auth_response));
 	  }
       if ((smtp_response[0] == '1' || smtp_response[0] == '2' || smtp_response[0] == '3') && smtp_response[3] == ' ') {
-	  if (*opt & ESMTP_AUTH)
-		SMTP_auth(sock, smtp_mode, name, password, auth_response);
+	  if (*opt & ESMTP_AUTH && name && password) /* name is defaulted to the login name... */
+		return SMTP_auth(sock, smtp_mode, name, password, auth_response);
 	  return SM_OK;
       }
       else if (smtp_response[3] != '-')
@@ -274,14 +321,18 @@ int SMTP_ehlo(int sock, char smtp_mode /**< 'S' for ESMTP or 'L' for LMTP */,
   return SM_UNRECOVERABLE;
 }
 
+static char *SMTP_fixup_address_literal(const char *addr /** input address (user@domain) */);
+
 int SMTP_from(int sock,
         char smtp_mode /** only for logging */,
-        const char *from /** envelope sender */,
-        const char *opts /** options for MAIL FROM:<>, will be concatenated to the command without additional blanks. NULL for no addition. */)
+        const char *_from /** envelope sender */,
+        const char *opts /** options for MAIL FROM:<>, needs to start with a blank, will be concatenated to the command. NULL for no addition. */)
 /**< send a "MAIL FROM:<\a from>" message to the SMTP listener */
 {
     int ok;
     char buf[MSGBUFSIZE];
+    char *from = SMTP_fixup_address_literal(_from);
+    if (!from) return SM_UNRECOVERABLE;
 
     if (from[0]=='<')
 	snprintf(buf, sizeof(buf), "MAIL FROM:%s", from);
@@ -293,23 +344,25 @@ int SMTP_from(int sock,
     if (outlevel >= O_MONITOR)
 	report(stdout, "%cMTP> %s\n", smtp_mode, buf);
     ok = SMTP_ok(sock, smtp_mode, TIMEOUT_MAIL);
+    xfree(from);
     return ok;
 }
 
-/** Send a "RCPT TO:" message with address \a to to the SMTP listener on socket \a sock */
-int SMTP_rcpt(int sock /** connected SMTP socket */,
-        char smtp_mode /** only for logging */,
-        const char *to /** destination address - bare address literals after the last @ on the line will be reformatted for RFC-5321 compliance*/)
+/** This function returns a copy of \a addr with the domain part
+ * reformatted as an IPv4 or IPv6 address literal, or NULL on failure.
+ * \returns NULL or a buffer that the caller must \ref free(). */
+static char *SMTP_fixup_address_literal(const char *addr /** input address (user@domain) */)
 {
-  int ok;
-  char *allocd = NULL;
-  const char *pfx = NULL;
-  const char *pos = NULL;
-  char *newlit = NULL;
+    const char *pfx = NULL;
+    const char *pos = NULL;
+    char *allocd = NULL;
 
-  if (1 == match_regex("@[0-9a-fA-Fx:.]+[[:space:]]*$", to))
-  {
-    pos = strrchr(to, '@');
+    char newlit[INET_ADDRSTRLEN + 1] = "";
+
+    if (1 != match_regex("@[0-9a-fA-Fx:.]+[[:space:]]*$", addr))
+        return xstrdup(addr);
+
+    pos = strrchr(addr, '@');
     ASSERT(pos);
     pos++;
     union {
@@ -336,38 +389,45 @@ int SMTP_rcpt(int sock /** connected SMTP socket */,
 	    // this will barf on 255.255.255.255 but that shan't be used as
 	    // smtphost anyways
       pfx = "[";
-      newlit = (char *)xmalloc(INET_ADDRSTRLEN + 1);
       inet_ntop(AF_INET, &in.in4, newlit, INET_ADDRSTRLEN);
     }
 #endif
-  }
 
-  if (pfx) {
-    if (pos < to || pos - to > INT_MAX) {
-      ok = SM_UNRECOVERABLE;
-      goto SMTP_rcpt_fail;
+    if (!pfx) return xstrdup(addr);
+
+    if (pos < addr || pos - addr > INT_MAX) {
+        return NULL;
     }
 
-    size_t needsize = pos - to + strlen(pfx) + 2 + strlen(newlit ? newlit : pos);
+    size_t needsize =
+        pos - addr + strlen(pfx) + 2 + strlen(*newlit ? newlit : pos);
     // what's past the @ is a literal IPv4 or IPv6 and needs [] around
     allocd = (char *)xmalloc(needsize);
-    int havesize = snprintf(allocd, needsize, "%-.*s%s%s]",
-	(int)(pos - to), to, pfx, newlit ? newlit : pos);
+    int havesize = snprintf(allocd, needsize, "%-.*s%s%s]", (int)(pos - addr), addr,
+                            pfx, *newlit ? newlit : pos);
     if (havesize <= 0 || (unsigned)havesize >= needsize) {
-      ok = SM_UNRECOVERABLE;
-      goto SMTP_rcpt_fail;
+        xfree(allocd);
+        return NULL;
     }
-    to = allocd;
-  }
+    return allocd;
+}
 
-  SockPrintf(sock,"RCPT TO:<%s>\r\n", to);
-  if (outlevel >= O_MONITOR)
-    report(stdout, "%cMTP> RCPT TO:<%s>\n", smtp_mode, to);
-  ok = SMTP_ok(sock, smtp_mode, TIMEOUT_RCPT);
-SMTP_rcpt_fail:
-  xfree(newlit);
-  xfree(allocd);
-  return ok;
+/** Send a "RCPT TO:" message with address \a to to the SMTP listener on socket \a sock */
+int SMTP_rcpt(int sock /** connected SMTP socket */,
+        char smtp_mode /** only for logging */,
+        const char *to /** destination address - bare address literals after the last @ on the line will be reformatted for RFC-5321 compliance*/)
+{
+    int ok;
+    char *newto = SMTP_fixup_address_literal(to);
+    if (!newto)
+      return SM_UNRECOVERABLE;
+
+    SockPrintf(sock, "RCPT TO:<%s>\r\n", newto);
+    if (outlevel >= O_MONITOR)
+      report(stdout, "%cMTP> RCPT TO:<%s>\n", smtp_mode, newto);
+    xfree(newto);
+    ok = SMTP_ok(sock, smtp_mode, TIMEOUT_RCPT);
+    return ok;
 }
 
 int SMTP_data(int sock, char smtp_mode /** only for logging */)
